;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 01_scheduler.lisp - Plánovač procesů
;;;

#|
Proces je reprezentováno seznamem:
(ret-stack seek-stack rslt-stack exec-stack bnd-stack)
|#

(defun make-process (ret seek rslt exec bdn)
  (list ret seek rslt exec bdn))

;; Nastaví proces jako běžící.
(defun run-process (process)
  (setf *ret* (first process)
        *seek* (second process)
        *rslt* (third process)
        *exec* (fourth process)
        *bnd* (fifth process)))

;; Vrátí běžící proces.
(defun running-process ()
  (make-process *ret* *seek* *rslt* *exec* *bnd*))

;; Frontu procesů připravených k vykonávání simulujeme pomocí dvou zásobníků.
;; Frontu libovolně rozdělíme na dvě části.
;; Zásobník *rprs-head* reprezentuje přední část fronty. Vrchol zásobníku se shoduje s hlavou fronty. 
;; Zásobník *rprs-tail* reprezentuje zadní část fronty. Vrchol zásobníku se shoduje s ocasem fronty.
;; (rprs je zkratka za Ready PRocesseS)

(defvar *rprs-head* nil)
(defvar *rprs-tail* nil)

;; Přeskládá procesy ze zadní části do přední.
(defun rearrange-ready-processes ()
  (when *rprs-tail*
    (push (pop *rprs-tail*) *rprs-head*)
    (rearrange-ready-processes)))

;; Uloží proces do fronty procesů připravených k běhu.
(defun insert-ready-process (process)
  (push process *rprs-tail*))

;; Odebere a vrátí proces z fronty připravených procesů.
(defun remove-ready-process ()
  (unless *rprs-head*
    (rearrange-ready-processes))
  (pop *rprs-head*))
     

;; Čas do spuštění plánovače:
(defvar *time*)

;; Maximální čas nepřerušeného vykonávání procesu:
(defvar *max-time*)
(setf *max-time* 100)

;; Vytvoří zahálčivý proces. 
(defun make-idle-process ()
  (make-process '() '() '() '(:lbl t :cjmp) *bnd*))
  
;; Dispečer:
(defun dispatcher ()
  (setf *time* (1+ (random *max-time*)))
  (run-process (remove-ready-process)))

;; Obsluha časovače:
(defun timer-handler ()
  (insert-ready-process (running-process))
  (dispatcher))

;; Slovo ukončující vedlejší proces:
(defprim :quit 
  (dispatcher))

;;
;; Spuštění programu.
;;

;; Potlačíme varování při změně definice funkce execute.
(setf dspec:*redefinition-action* :quiet)

(defun execute (&rest code)
  (let ((*ret* '())
        (*seek* '())
        (*rslt* '())
        (*exec* code) 
        (*time* 0)
        (*rprs-head* (list (make-idle-process)))
        (*rprs-tail* '()))
    (loop 
       (cond 
        ((null *exec*)
         (return))
        ((= *time* 0)
         (timer-handler))
        (t
         (exec-elem (pop *exec*))
         (decf *time*))))
    (pop *rslt*)))


(setf dspec:*redefinition-action* :warn)

;; Slovo na vytvoření vedlejšího procesu: 
(defprim :prsub
  (let ((sub (pop *rslt*))
        (arg (pop *rslt*)))
    (insert-ready-process (make-process '() '() (list arg) sub *bnd*))))
 


;; Aktivní čekání. Opakovaně vykonává podprogram na vrcholu datového zásobníku.
;; Vykonání podprogramu musí způsobit efekt: (-- val) 
;; Opakování končí, když je hodnota val Pravda.
;;
;; (sub --)
(execute
 :def :await :lbl :dup :clsub :not :cjmp :drop)


#|
; Hlavní proces bude čekat na skončení vedlejšího procesu:

nil end? :bind 
1 (:print t end? :set! :exit) :prsub 
(end? :val :ret) :await 
:unbind
|#





