;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 08_barrier.lisp - Bariéry
;;;

;; Nastavení odsazení operátoru while
(editor:setup-indent "while" 1)


;; Dále si zavedeme několik užitečných operátorů.

;; let*

(def-scheme-macro let* (bindings &rest body)
  (if bindings
      `(let (,(car bindings))
         (let* ,(cdr bindings)
           ,@body))
    `(begin ,@body)))

#|
;; Scheme

(let* 
    ((a 1)
     (b (+ a 1)))
  (print a)
  (print b))
|#


;; Logaritmus, mocnění a zaokrouhlování nahoru
;; (number base -- log number o základu base)
(defprim :log 
  (let ((base (pop *rslt*))
        (number (pop *rslt*)))
    (push (log number base) *rslt*)))

#|
8 2 :log
|#

(execute '(:log :ret) 'log :bind)

#|
;; Scheme
(log 8 2)
|#

;; Mocnina
(compiled-evaluate '(define expt (lambda (base power)
                                   (if (= power 0)
                                       1
                                     (* (expt base (- power 1)) base)))))

#|
;; Scheme
(expt 2 3)
|#

;; (number -- číslo number zaokrouhlené nahoru)
(defprim :ceiling 
  (push (ceiling (pop *rslt*)) *rslt*))

#|
3.1 :ceiling
|#

(execute '(:ceiling :ret) 'ceiling :bind)

#|
;; Scheme

(ceiling 3.1)
|#

;; Makro implikace: ->
(def-scheme-macro -> (&rest args)
  (if (cdr (cdr args))
      `(-> ,(car args) (-> ,@(cdr args)))
    `(or (not ,(car args)) ,(cadr args))))

#|
;; Scheme
(-> (= 1 1) (= 1 2))
|#

;; Atomická operace Fetch-and-add
(def-scheme-macro fa (var num)
  (let ((res (gensym "RES")))
    `(atomic 
      (let ((,res ,var))
        (set! ,var (+ ,res ,num))
        ,res))))

#|
;; Scheme
(let ((n 0))
  (co-dotimes (i 10)
    (fa n 1))
  n)
|#

;; Nyní se už můžeme věnovat bariérám.

;; Makro with-barrier
;; Syntax:
;;   (with-barrier (count) . body)
;;
;; Číslo count udává počet procesů, které chceme bariérou synchronizovat.
;; V těle můžeme použít funkci barrier, která implementuje bariéru.
;; Implementace bariéry je podobná bariéře s počítadlem z přednášky
;; až na to, že používá místo aktivního čekání semafory.
;; Příklad použití je za definicí makra.

(def-scheme-macro with-barrier (count-part &rest body)
  (let ((count-form (car count-part))
        (turnstile (gensym "TURNSTILE"))
        (turnstile2 (gensym "TURNSTILE2"))
        (lock (gensym "LOCK"))
        (count (gensym "COUNT"))
        (barrier-count (gensym "BARRIER-COUNT")))
    `(let ((,count ,count-form)
           (,turnstile (sem 0))
           (,turnstile2 (sem 1))
           (,lock (lock))
           (,barrier-count 0))
       (let ((barrier 
              (lambda ()
                (with-lock (,lock)
                  (inc! ,barrier-count)
                  (when (= ,barrier-count ,count)
                    (p ,turnstile2)
                    (v ,turnstile)))
                (p ,turnstile)
                (v ,turnstile)
                (with-lock (,lock)
                  (dec! ,barrier-count)
                  (when (= ,barrier-count 0)
                    (p ,turnstile)
                    (v ,turnstile2)))
                (p ,turnstile2)
                (v ,turnstile2))))
         ,@body))))

;; Nastavení odsazení operátoru with-barrier
(editor:setup-indent "with-barrier" 1)

#|
;; Scheme
(with-barrier (5)
  (co-dotimes (i 5)
    (dotimes (j 10)
      (print j)
      (barrier))))
|#

;; Makro barrierd-co-dotimes kombinuje makra co-dotimes a barrier.

(def-scheme-macro barrierd-co-dotimes (iter-part &rest body)
  (let ((var (car iter-part))
        (count-form (cadr iter-part))
        (count (gensym "COUNT")))
    `(let ((,count ,count-form))
       (with-barrier (,count)
         (co-dotimes (,var ,count)
           ,@body)))))

;; Nastavení odsazení operátoru barrierd-co-dotimes
(editor:setup-indent "barrierd-co-dotimes" 1)

#|
;; Scheme
(barrierd-co-dotimes (i 5)
  (dotimes (j 10)
    (print j)
    (barrier)))
|#


;; Procedury na práci s vlakami.
(compiled-evaluate '(define flag-clear!
                            (lambda (flags i)
                              (vset! flags i nil))))

(compiled-evaluate '(define flag-set!
                            (lambda (flags i)
                              (vset! flags i t))))

(compiled-evaluate '(define flag-set? vref))

(compiled-evaluate '(define wait-and-clear! 
                            (lambda (flags i)
                              (await (flag-set? flags i))
                              (flag-clear! flags i))))

(compiled-evaluate '(define wait-and-set! 
                            (lambda (flags i)
                              (await (not (flag-set? flags i)))
                              (flag-set! flags i))))




